library ac_dart.auth;
import 'dart:io';
import 'dart:async';

import 'util.dart';
import 'package:option/option.dart';
import 'package:either/either.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';

import 'package:logging/logging.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_auth/shelf_auth.dart';
export 'package:shelf/shelf.dart' show Request, Response, Handler, Middleware;

Logger _log = new Logger('ac_dart.auth');

const AC_DART_SESSION_PARAM = 'atlassian-acdart-session';
const AC_DART_SESSION_PARAM_HOST_KEY = '$AC_DART_SESSION_PARAM-hostKey';
const AC_DART_SESSION_PARAM_HOST_URL = '$AC_DART_SESSION_PARAM-hostUrl';
const AC_DART_SESSION_PARAM_USERNAME = '$AC_DART_SESSION_PARAM-userName';
const AC_DART_SESSION_PARAM_TOKEN = '$AC_DART_SESSION_PARAM-token';
const AC_DART_SESSION_RESPONSE_HEADER = 'X-$AC_DART_SESSION_PARAM_TOKEN';


abstract class Authenticator {
  dynamic authenticate(Request request, Handler innerHandler);
  
  Middleware get middleware;
}

abstract class AbstractHostServiceAuthenticator implements Authenticator {
  final AcHostService _hostService;
  final String queryParamName;
  

  AbstractHostServiceAuthenticator(this._hostService, this.queryParamName);
  
  Middleware get middleware => _handler;
  
  Handler _handler(Handler innerHandler) {
    return (Request request) {
      return authenticate(request, innerHandler);
    };
  }
  
  
  @override
  dynamic authenticate(Request request, Handler innerHandler) {

    _log.finer('authenticating ${request.method} ${request.requestedUri}');

    final Option<String> jwtTokenOpt = 
        new Option(request.requestedUri.queryParameters[queryParamName])
      .orElse(_fromAuthHeader(request.headers[HttpHeaders.AUTHORIZATION]));
    
    return jwtTokenOpt.map((jwtToken) {
      _log.finest('token is $jwtToken');

      final Future<Either> authResultFuture = 
          authenticateRequest(jwtToken, request);
      
      return authResultFuture.then((Either either) {
        if (either.isRight()) {
          _log.finer('request successfully authenticated ${request.method} '
              '${request.requestedUri}');
          final AcSessionContext session = either.right().get();
          return _invokeHandlerWithContext(session, request, innerHandler);
        }
        else {
          _log.finer('request failed authentication ${request.method} '
              '${request.requestedUri}');
          AuthenticationFailure authenticationFailure = either.left().get();
          return new Response(authenticationFailure.statusCode, 
              body: authenticationFailure.message);
        }
      });
    })
    .getOrElse(() {
      _log.finer('request failed authentication - no jwt in request: '
          '${request.method} ${request.requestedUri}');
      return new Response(HttpStatus.UNAUTHORIZED, body: 
        'no jwt token found in request');
    });
  }

  Future<Response> _invokeHandlerWithContext(AcSessionContext session, 
      Request request, Handler innerHandler) {
    return _hostService.findRegistrationByHostKey(session.hostKey).then((regOpt) =>
        regOpt.map((hostRego) { 
          final newSessionToken = 
              _hostService.createAddonSessionToken(session.hostKey, session.user);
          final newRequest = _copyRequestWith(request, extraParamsDelta: {
            AC_DART_SESSION_PARAM_HOST_KEY: session.hostKey,
            // TODO: maybe shouldn't add host url in the session. Overkill. Leave it to moustache middleware??
            AC_DART_SESSION_PARAM_HOST_URL: hostRego.hostDetails.baseUrl.toString(),
            AC_DART_SESSION_PARAM_USERNAME: session.user,
            AC_DART_SESSION_PARAM_TOKEN: newSessionToken
          });
          return syncFuture(() => innerHandler(newRequest)).then((r) =>
              _responseWithSessionToken(r, newSessionToken));
        })
        .getOrElse(new Response.internalServerError(body: "no host registration found")));
  }
  
  Request _copyRequestWith(Request r, { Map<String, String> extraParamsDelta }) {
    // TODO: currently shelf does not support any parameters beyond the request
    // headers. However there is an issue for this so hopefully soon. For now
    // whack em in the header.
    // TODO: also there has been talk of adding a copyWith method to request
    // which we would use here if it existed
    
    final headers = new Map<String, String>()
        ..addAll(r.headers)
        ..addAll(extraParamsDelta);
    
    return new Request(r.method, r.requestedUri,
        protocolVersion: r.protocolVersion, headers: headers, 
        url: r.url, scriptName: r.scriptName,
        body: r.read());
  }
  
  Response _copyResponseWith(Response r, { Map<String, String> headersDelta }) {
    // TODO: there has been talk of adding a copyWith method to Response
    // which we would use here if it existed
    
    final headers = new Map<String, String>()
        ..addAll(r.headers)
        ..addAll(headersDelta);
    
    return new Response(r.statusCode, body: r.read(), headers: headers);
  }
  
  Response _responseWithSessionToken(Response r, String sesionToken) {
    return _copyResponseWith(r, headersDelta: 
      { AC_DART_SESSION_RESPONSE_HEADER: sesionToken });    
  }

  Option<String> _fromAuthHeader(String authHeader) {
    List<String> authHeaders = authHeader == null ? [] : authHeader.split(',');
    final firstJwtHeader = authHeaders.firstWhere((authHeader) => 
        authHeader.startsWith('JWT'), orElse: () => '');
    
    final authTypeCredentialPair = firstJwtHeader.split(' ');
    if (authTypeCredentialPair.length != 2) {
      return const None();
    }
    
    return new Some(authTypeCredentialPair[1]);
  }

  Future<Either<AuthenticationFailure, AcSessionContext>> authenticateRequest(
        String jwtToken, Request request);
}

class JwtAuthenticator extends AbstractHostServiceAuthenticator {
  JwtAuthenticator(AcHostService _hostService)
      : super(_hostService, 'jwt');

  @override
  Future<Either<AuthenticationFailure, AcSessionContext>> authenticateRequest(
      String jwtToken, Request request) =>
          _hostService.authenticateHostRequest(
              jwtToken, request.requestedUri, request.method);  
}

class SessionTokenAuthenticator extends AbstractHostServiceAuthenticator {
  SessionTokenAuthenticator(AcHostService _hostService)
      : super(_hostService, 'acpt');

  @override
  Future<Either<AuthenticationFailure, AcSessionContext>> authenticateRequest(
      String jwtToken, Request request) =>
          _hostService.authenticateAddonSessionToken(jwtToken);

}

AcSessionContext getSessionContext(Request request) {
  final hostKey = request.headers[AC_DART_SESSION_PARAM_HOST_KEY];
  final username = request.headers[AC_DART_SESSION_PARAM_USERNAME];
  
  if (hostKey == null || username == null) {
    throw new ArgumentError('missing or invalid session context in request');
  }
  
  return new AcSessionContext(username, hostKey);
}

class HcPrincipal extends Principal {
  final int id;
  
  HcPrincipal(this.id, String name) : super(name);
  
}