library ac_dart.auth.test;

import 'package:atlassian_connect_shelf/src/auth.dart';
import 'package:unittest/unittest.dart';
import 'package:mock/mock.dart';
import 'dart:async';

import 'package:atlassian_connect_jwt/atlassian_connect_jwt.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';
import 'package:atlassian_connect_host/host_model.dart';
import 'package:logging/logging.dart';
import 'package:either/either.dart';
import 'package:option/option.dart';

class _MockAcHostService extends Mock implements AcHostService {
  noSuchMethod(i) => super.noSuchMethod(i);
}

//mocking: typedef Handler(Request request);
abstract class HandlerClass {
  dynamic call(Request request) { return null; }
}

class MockHandler extends Mock implements HandlerClass {
  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}

void _initLogging() {
  Logger.root.level = Level.WARNING;
  Logger.root.onRecord.listen((LogRecord rec) {
    print('${rec.message}');
  });
}

void main() {
  _initLogging();

  MockHandler handler;

  _MockAcHostService hostService;

  Request request;

  Authenticator authenticator;


  final String hostKey = "jira:ae390d29-31b2-4c12-a719-9df64e3e92b7";
  final String user = 'fred';
  final String sharedSecret = '3ab90b11-d7bd-4097-958f-01b7ac4e985f';
  final String sessionToken = 'la tokena';

  final hostRego = (new AcHostRegistration.build(
      sharedSecret: "$sharedSecret",
      hostDetails: (new AcHostDetails.build(
          key: "$hostKey",
          baseUrl: Uri.parse('http://localhost:8080/'),
          productType: ProductType.jira,
          publicKey: "pubby"))));


  setUp(() {
    hostService = new _MockAcHostService();
    handler = new MockHandler();
    handler.when(callsTo('call'))
      .alwaysReturn(new Response.ok("woohoo"));

    hostService.when(callsTo('findRegistrationByHostKey'))
        .alwaysReturn(new Future.value(new Some(hostRego)));

    hostService.when(callsTo('createAddonSessionToken')).thenReturn(sessionToken);
  });

  void mockRequest(Uri uri) {
    request = new Request("GET", uri,
        protocolVersion: "1.1",
        headers: {});
  }

  Future<Response> authenticate(Uri uri) {
    mockRequest(uri);
    return new Future.value(authenticator.authenticate(request, handler));
  }

  void _expect(Uri uri, int expectedResponseStatus) {
    final authResultFuture = authenticate(uri);
    authResultFuture.then((Response response) {
      expect(response.statusCode, equals(expectedResponseStatus));
    });
    expect(authResultFuture, completes);
  }

  void _expectResponse(Uri uri, void check(Response response)) {
    final authResultFuture = authenticate(uri);
    authResultFuture.then((Response result) {
      check(result);
    });
    expect(authResultFuture, completes);
  }

  void runAuthTests(Uri uriNoJwt, Uri uriInvalidJwt, Uri uriWithValidJwt) {
    group('[no jwt]', () {
      test('response is 401 when no jwt in request', () {
        _expect(uriNoJwt, 401);
      });
    });

    group('[auth failure]', () {
      test('response is 401 when invalid jwt in request', () {
        _expect(uriInvalidJwt, 401);
      });
    });

    group('[auth success]', () {

      test('calls provided handler when valid jwt in request', () {
        _expectResponse(uriWithValidJwt, (_) {
          handler.calls('call').verify(happenedOnce);
        });
      });

      test('calls provided handler with session token in request', () {
        _expectResponse(uriWithValidJwt, (_) {
          handler.calls('call',
              new HasHeaderValue(AC_DART_SESSION_PARAM_TOKEN, isNotNull))
              .verify(happenedOnce);
        });
      });

      test('calls provided handler with hostkey in request', () {
        _expectResponse(uriWithValidJwt, (_) {
          handler.calls('call',
              new HasHeaderValue(AC_DART_SESSION_PARAM_HOST_KEY, equals(hostKey)))
              .verify(happenedOnce);
        });
      });

      test('calls provided handler with username in request', () {
        _expectResponse(uriWithValidJwt, (_) {
          handler.calls('call',
              new HasHeaderValue(AC_DART_SESSION_PARAM_USERNAME, equals(user)))
              .verify(happenedOnce);
        });
      });

      test('returns provided handlers response', () {
        _expect(uriWithValidJwt, 200);
      });

      test('response has token in header', () {
        _expectResponse(uriWithValidJwt, (response) {
          expect(response.headers, containsPair(AC_DART_SESSION_RESPONSE_HEADER,
                                                sessionToken));
        });
      });
    });

  }


  group('[jwt authenticator]', () {
    final uriStrNoJwt = "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&xdm_c=channel-servlet-hello-world&xdm_p=1";
    final uriNoJwt = Uri.parse(uriStrNoJwt);

    final String invalidJwtStr = r'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjEzOTQxNjYzNjcsInN1YiI6ImFkbWluIiwiaXNzIjoiamlyYTphZTM5MGQyOS0zMWIyLTRjMTItYTcxOS05ZGY2NGUzZTkyYjciLCJxc2giOiJlYjRlYzVmMDk4MGQwMWRhMzA2ZGI4OWZhNTdkZDE2MzU3NDY2NzQzNmRkNzIwZDVkOWM1Mjc5YzViNDVmN2E4IiwiaWF0IjoxMzk0MTY2MTg3fQ.bR8Z0MIguOg6xgHiav0quun8kTqXzNUKMKym-PFjZvc';
    final uriInvalidJwt = Uri.parse("http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&xdm_c=channel-servlet-hello-world&xdm_p=1&jwt=${invalidJwtStr}");

    String jwtToken = createProductHostToken(uriNoJwt.path, 'GET',
    uriNoJwt.queryParameters, sharedSecret, hostKey);

    final uriWithValidJwt = Uri.parse('$uriStrNoJwt&jwt=$jwtToken');


    setUp(() {
      authenticator = new JwtAuthenticator(hostService);

      hostService.when(callsTo('authenticateHostRequest', anything, equals(uriWithValidJwt)))
        .alwaysReturn(new Future.value(new Right(new AcSessionContext(user, hostKey))));

      hostService.when(callsTo('authenticateHostRequest', anything, equals(uriInvalidJwt)))
        .alwaysReturn(new Future.value(new Left(new AuthenticationFailure(401,
            new ConstraintViolations('oops', new Set()..add(new ConstraintViolation('bugger')))))));
    });

    runAuthTests(uriNoJwt, uriInvalidJwt, uriWithValidJwt);
  });


  group('[session token authenticator]', () {
    const String addonSecret = 'el secreto';
    const String addonKey = 'la addona';

    final uriStrNoJwt = "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&"
        + "user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&"
        + "xdm_c=channel-servlet-hello-world&xdm_p=1";
    final uriNoJwt = Uri.parse(uriStrNoJwt);

    final String invalidJwtStr = r'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9'
        + '.eyJleHAiOjEzOTQxNjYzNjcsInN1YiI6ImFkbWluIiwiaXNzIjoiamlyYTphZTM5MGQy'
        + 'OS0zMWIyLTRjMTItYTcxOS05ZGY2NGUzZTkyYjciLCJxc2giOiJlYjRlYzVmMDk4MG'
        + 'QwMWRhMzA2ZGI4OWZhNTdkZDE2MzU3NDY2NzQzNmRkNzIwZDVkOWM1Mjc5YzViNDVm'
        + 'N2E4IiwiaWF0IjoxMzk0MTY2MTg3fQ.bR8Z0MIguOg6xgHiav0quun8kTqXzNUKMKym-PFjZvc';
    final uriInvalidJwt = Uri.parse(
        "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&"
        + "loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&"
        + "xdm_c=channel-servlet-hello-world&xdm_p=1&jwt=${invalidJwtStr}");

    final String validJwtStr = createAddonSessionToken(addonSecret, addonKey,
       hostKey, user);

    final uriWithValidJwt = Uri.parse('$uriStrNoJwt&acpt=$validJwtStr');


    setUp(() {
      authenticator = new SessionTokenAuthenticator(hostService);

      hostService.when(callsTo('authenticateAddonSessionToken', validJwtStr))
          .alwaysReturn(new Future.value(new Right(new AcSessionContext(user,
              hostKey))));

      hostService.when(callsTo('authenticateAddonSessionToken', invalidJwtStr))
          .alwaysReturn(new Future.value(new Left(new AuthenticationFailure(401,
              new ConstraintViolations('oops', new Set()..add(
                  new ConstraintViolation('bugger')))))));
    });

    runAuthTests(uriNoJwt, uriInvalidJwt, uriWithValidJwt);

  });
}


class HasHeaderValue extends CustomMatcher {
  final String headerName;
  HasHeaderValue(String headerName, Matcher matcher)
    : this.headerName = headerName,
      super("Request with header whose key is '$headerName' and value that is",
          "headers", matcher);

  String featureValueOf(Request actual) => actual.headers[headerName];
}